import os
import threading
import re
import math
import uuid
import zipfile
import requests
import urllib3.exceptions
from retrying import retry
from tqdm import tqdm
import arrow
from .preprocess import Project
from . import __version__


def retry_on_http_error(exception):
    valid_exceptions = [
        requests.exceptions.HTTPError,
        requests.exceptions.ConnectionError,
        urllib3.exceptions.HTTPError,
    ]
    return any([isinstance(exception, e) for e in valid_exceptions])


ANNOTATIONS_IMPORT_DEFAULT_CHUNK_SIZE = 100000


class Client:
    """Client for communicating with MD.ai backend API.
    Communication is via user access tokens (in MD.ai Hub, Settings -> User Access Tokens).
    """

    def __init__(self, domain="public.md.ai", access_token=None):
        domain_match = re.match(r"^\w+\.md\.ai$", domain)
        dev_domain_match = re.match(r"^\w+\.mdai.dev(:\d+)?$", domain)
        if not domain_match and not dev_domain_match:
            raise ValueError(f"domain {domain} is invalid: should be format *.md.ai")

        self.domain = domain
        self.access_token = access_token
        self.session = requests.Session()
        self._test_endpoint()

    def project(
        self,
        project_id,
        dataset_id=None,
        label_group_id=None,
        path=".",
        force_download=False,
        annotations_only=False,
    ):
        """Initializes Project class given project id.

        Arguments:
            project_id: hash ID of project
            dataset_id: hash ID of dataset to scope to (optional - default `None`)
            label_group_id: hash ID of the label group to scope to (optional - default `None`)
            path: directory used for data (optional - default `"."`)
            force_download: if `True`, ignores possible existing data in `path` (optional - default `False`)
            annotations_only: if `True`, downloads annotations only (optional - default `False`)
        """
        if path == ".":
            print("Using working directory for data.")
        else:
            os.makedirs(path, exist_ok=True)
            print(f"Using path '{path}' for data.")

        data_manager_kwargs = {
            "domain": self.domain,
            "project_id": project_id,
            "dataset_id": dataset_id,
            "label_group_id": label_group_id,
            "path": path,
            "session": self.session,
            "headers": self._create_headers(),
            "force_download": force_download,
        }

        annotations_data_manager = ProjectDataManager("annotations", **data_manager_kwargs)
        annotations_data_manager.create_data_export_job()
        if not annotations_only:
            images_data_manager = ProjectDataManager("images", **data_manager_kwargs)
            images_data_manager.create_data_export_job()

        annotations_data_manager.wait_until_ready()
        if not annotations_only:
            images_data_manager.wait_until_ready()
            p = Project(
                annotations_fp=annotations_data_manager.data_path,
                images_dir=images_data_manager.data_path,
            )
            return p
        else:
            print("No project created. Downloaded annotations only.")
            return None

    def download_model_outputs(
        self, project_id, dataset_id=None, model_id=None, path=".", force_download=False
    ):
        """Downloads model outputs given project_id.

        Arguments:
            project_id: hash ID of project
            dataset_id: hash ID of dataset (optional - default `None`)
            model_id: hash ID of the model (optional - default `None`)
            path: directory used for data (optional - default `"."`)
            force_download: if `True`, ignores possible existing data in `path` (optional - default `False`)
        """
        if path == ".":
            print("Using working directory for model outputs.")
        else:
            os.makedirs(path, exist_ok=True)
            print(f"Using path '{path}' for data.")

        data_manager_kwargs = {
            "domain": self.domain,
            "project_id": project_id,
            "dataset_id": dataset_id,
            "model_id": model_id,
            "path": path,
            "session": self.session,
            "headers": self._create_headers(),
            "force_download": force_download,
        }

        model_outputs_manager = ProjectDataManager("model-outputs", **data_manager_kwargs)
        model_outputs_manager.create_data_export_job()
        model_outputs_manager.wait_until_ready()
        return None

    def load_model_annotations(self):
        """Deprecated method: use `import_annotations` instead.
        """
        print("Deprecated method: use `import_annotations` instead.")

    def import_annotations(
        self, annotations, project_id, dataset_id, chunk_size=ANNOTATIONS_IMPORT_DEFAULT_CHUNK_SIZE,
    ):
        """Import annotations into project.
        For example, this method can be used to load machine learning model results into project as
        annotations, or quickly populate metadata labels.

        Arguments:
            project_id: hash ID of project.
            dataset_id: hash ID of dataset.
            annotations: list of annotations to load.
            chunk_size: number of annotations to load as a chunk.
        """
        if not annotations:
            print("No annotations provided.")
        if not project_id:
            print("project_id is required.")
        if not dataset_id:
            print("dataset_id is required.")

        num_chunks = math.ceil(len(annotations) / chunk_size)

        if num_chunks > 1:
            print(f"Importing {len(annotations)} total annotations in {num_chunks} chunks...")

        failed_annotations = []

        for i in range(num_chunks):
            if num_chunks > 1:
                print(f"Chunk {i+1}...")

            start = i * chunk_size
            end = (i + 1) * chunk_size
            annotations_chunk = annotations[start:end]

            manager = AnnotationsImportManager(
                annotations=annotations_chunk,
                project_id=project_id,
                dataset_id=dataset_id,
                session=self.session,
                domain=self.domain,
                headers=self._create_headers(),
            )
            manager.create_job()
            manager.wait_until_ready()

            for failed_annotation in manager.failed_annotations:
                # add start index since returned index is for chunk
                failed_annotation["index"] += start
                failed_annotations.append(failed_annotation)

        if num_chunks > 1:
            num_failed = len(failed_annotations)
            print(
                f"Successfully imported {len(annotations) - num_failed} / {len(annotations)}"
                + f" total annotations into project {project_id}."
            )

        return failed_annotations

    def _create_headers(self):
        headers = {}
        if self.access_token:
            headers["x-access-token"] = self.access_token
        return headers

    def _test_endpoint(self):
        """Checks endpoint for validity and authorization.
        """
        test_endpoint = f"https://{self.domain}/api/test"
        r = self.session.get(test_endpoint, headers=self._create_headers())
        if r.status_code == 200:
            print(f"Successfully authenticated to {self.domain}.")
        else:
            raise Exception("Authorization error. Make sure your access token is valid.")

    @retry(
        retry_on_exception=retry_on_http_error,
        wait_exponential_multiplier=100,
        wait_exponential_max=1000,
        stop_max_attempt_number=10,
    )
    def _gql(self, query, variables=None):
        """Executes GraphQL query.
        """
        gql_endpoint = f"https://{self.domain}/api/graphql"
        headers = self._create_headers()
        headers["Accept"] = "application/json"
        headers["Content-Type"] = "application/json"
        headers["apollographql-client-name"] = ("mdai-client-py",)
        headers["apollographql-client-version"] = __version__

        data = {"query": query, "variables": variables}
        r = self.session.post(gql_endpoint, headers=headers, json=data)
        if r.status_code != 200:
            r.raise_for_status()

        body = r.json()
        data = body["data"] if "data" in body else None
        errors = body["errors"] if "errors" in body else None

        return data, errors


class ProjectDataManager:
    """Manager for project data exports and downloads.
    """

    def __init__(
        self,
        data_type,
        domain=None,
        project_id=None,
        dataset_id=None,
        label_group_id=None,
        model_id=None,
        path=".",
        session=None,
        headers=None,
        force_download=False,
    ):
        if data_type not in ["images", "annotations", "model-outputs"]:
            raise ValueError("data_type must be 'images' or 'annotations' or 'model-outputs'.")
        if not domain:
            raise ValueError("domain is not specified.")
        if not project_id:
            raise ValueError("project_id is not specified.")
        if not os.path.exists(path):
            raise OSError(f"Path '{path}' does not exist.")

        self.data_type = data_type
        self.force_download = force_download

        self.domain = domain
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.label_group_id = label_group_id
        self.model_id = model_id
        self.path = path
        if session and isinstance(session, requests.Session):
            self.session = session
        else:
            self.session = requests.Session()
        self.headers = headers

        # path for downloaded data
        self.data_path = None
        # ready threading event
        self._ready = threading.Event()

    def create_data_export_job(self):
        """Create data export job through MD.ai API.
        This is an async operation. Status code of 202 indicates successful creation of job.
        """
        endpoint = f"https://{self.domain}/api/data-export/{self.data_type}"
        params = self._get_data_export_params()
        r = self.session.post(endpoint, json=params, headers=self.headers)
        if r.status_code == 202:
            msg = f"Preparing {self.data_type} export for project {self.project_id}..."
            print(msg.ljust(100))
            self._check_data_export_job_progress()
        else:
            if r.status_code == 401:
                msg = (
                    f"Project {self.project_id} at domain {self.domain}"
                    + " does not exist or you do not have sufficient permissions for access."
                )
                print(msg)
            self._on_data_export_job_error()

    def wait_until_ready(self):
        self._ready.wait()

    def _get_data_export_params(self):
        if self.data_type == "images":
            params = {
                "projectHashId": self.project_id,
                "datasetHashId": self.dataset_id,
                "exportFormat": "zip",
            }
        elif self.data_type == "annotations":
            # TODO: restrict to assigned labelgroup
            params = {
                "projectHashId": self.project_id,
                "datasetHashId": self.dataset_id,
                "labelGroupHashId": self.label_group_id,
                "exportFormat": "json",
            }
        elif self.data_type == "model-outputs":
            params = {
                "projectHashId": self.project_id,
                "datasetHashId": self.dataset_id,
                "modelHashId": self.model_id,
                "exportFormat": "json",
            }
        return params

    @retry(
        retry_on_exception=retry_on_http_error,
        wait_exponential_multiplier=100,
        wait_exponential_max=1000,
        stop_max_attempt_number=10,
    )
    def _check_data_export_job_progress(self):
        """Poll for data export job progress.
        """
        endpoint = f"https://{self.domain}/api/data-export/{self.data_type}/progress"
        params = self._get_data_export_params()
        r = self.session.post(endpoint, json=params, headers=self.headers)
        if r.status_code != 200:
            r.raise_for_status()

        try:
            body = r.json()
            status = body["status"]
        except (TypeError, KeyError):
            self._on_data_export_job_error()
            return

        if status == "done":
            self._on_data_export_job_done()

        elif status == "error":
            self._on_data_export_job_error()

        elif status == "running":
            try:
                progress = int(body["progress"])
            except (TypeError, ValueError):
                progress = 0
            try:
                time_remaining = int(body["timeRemaining"])
            except (TypeError, ValueError):
                time_remaining = 0

            # print formatted progress info
            if time_remaining > 45:
                time_remaining_fmt = (
                    arrow.now().shift(seconds=time_remaining).humanize(only_distance=True)
                )
            else:
                # arrow humanizes <= 45 to 'in seconds' or 'just now',
                # so we will opt to be explicit instead.
                time_remaining_fmt = f"{time_remaining} seconds"
            end_char = "\r" if progress < 100 else "\n"
            msg = (
                f"Exporting {self.data_type} for project {self.project_id}..."
                + f"{progress}% (time remaining: {time_remaining_fmt})."
            )
            print(msg.ljust(100), end=end_char, flush=True)

            # run progress check at 1s intervals so long as status == 'running'
            t = threading.Timer(1.0, self._check_data_export_job_progress)
            t.start()

    @retry(
        retry_on_exception=retry_on_http_error,
        wait_exponential_multiplier=100,
        wait_exponential_max=1000,
        stop_max_attempt_number=10,
    )
    def _on_data_export_job_done(self):
        endpoint = f"https://{self.domain}/api/data-export/{self.data_type}/done"
        params = self._get_data_export_params()
        r = self.session.post(endpoint, json=params, headers=self.headers)
        if r.status_code != 200:
            r.raise_for_status()

        try:
            file_keys = r.json()["fileKeys"]

            if file_keys:
                data_path = self._get_data_path(file_keys)
                if self.force_download or not os.path.exists(data_path):
                    # download in separate thread
                    t = threading.Thread(target=self._download_files, args=(file_keys,))
                    t.start()
                else:
                    # use existing data
                    self.data_path = data_path
                    print(f"Using cached {self.data_type} data for project {self.project_id}.")
                    # fire ready threading.Event
                    self._ready.set()
        except (TypeError, KeyError):
            self._on_data_export_job_error()

    @retry(
        retry_on_exception=retry_on_http_error,
        wait_exponential_multiplier=100,
        wait_exponential_max=1000,
        stop_max_attempt_number=10,
    )
    def _on_data_export_job_error(self):
        endpoint = f"https://{self.domain}/api/data-export/{self.data_type}/error"
        params = self._get_data_export_params()
        r = self.session.post(endpoint, json=params, headers=self.headers)
        if r.status_code != 200:
            r.raise_for_status()
        print(f"Error exporting {self.data_type} for project {self.project_id}.")
        # fire ready threading.Event
        self._ready.set()

    def _get_data_path(self, file_keys):
        if self.data_type == "images":
            # should be folder for zip file:
            # xxxx.zip -> xxxx/
            # xxxx_part1of3.zip -> xxxx/
            images_dir = re.sub(r"(_part\d+of\d+)?\.\S+$", "", file_keys[0])
            return os.path.join(self.path, images_dir)
        elif self.data_type == "annotations":
            # annotations export will be single file
            annotations_fp = file_keys[0]
            return os.path.join(self.path, annotations_fp)
        elif self.data_type == "model-outputs":
            # model outputs export will be single file
            model_outputs_fp = file_keys[0]
            return os.path.join(self.path, model_outputs_fp)

    def _download_files(self, file_keys):
        """Downloads exported files.
        """
        try:
            for file_key in file_keys:
                print(f"Downloading file: {file_key}")
                filepath = os.path.join(self.path, file_key)

                key = requests.utils.quote(file_key)
                dl_session_id = str(uuid.uuid4())

                # request download token
                url = f"https://{self.domain}/api/data-export/download-request"
                data = {"key": key, "sessionId": dl_session_id}
                r = requests.post(url, json=data, headers=self.headers)
                dl_token = r.json().get("token")

                # download file
                # stream response so we can display progress bar
                params = {"token": dl_token, "sessionId": dl_session_id}
                url = f"https://{self.domain}/api/data-export/download/{key}"
                r = requests.get(url, params=params, headers=self.headers, stream=True)

                # total size in bytes
                total_size = int(r.headers.get("content-length", 0))
                block_size = 32 * 1024
                wrote = 0
                with open(filepath, "wb") as f:
                    with tqdm(
                        total=total_size, unit="B", unit_scale=True, unit_divisor=1024
                    ) as pbar:
                        for chunk in r.iter_content(block_size):
                            f.write(chunk)
                            wrote = wrote + len(chunk)
                            pbar.update(block_size)
                if total_size != 0 and wrote != total_size:
                    raise IOError(f"Error downloading file {file_key}.")

                if self.data_type == "images":
                    # unzip archive
                    print(f"Extracting archive: {file_key}")
                    with zipfile.ZipFile(filepath, "r") as f:
                        f.extractall(self.path)

            self.data_path = self._get_data_path(file_keys)

            print(f"Success: {self.data_type} data for project {self.project_id} ready.")
        except Exception:
            print(f"Error downloading {self.data_type} data for project {self.project_id}.")

        # fire ready threading.Event
        self._ready.set()


class AnnotationsImportManager:
    """Manager for importing annotations.
    """

    def __init__(
        self,
        annotations=None,
        project_id=None,
        dataset_id=None,
        session=None,
        domain=None,
        headers=None,
    ):
        if not domain:
            raise ValueError("domain is not specified.")
        if not project_id:
            raise ValueError("project_id is not specified.")

        self.annotations = annotations
        self.project_id = project_id
        self.dataset_id = dataset_id
        if session and isinstance(session, requests.Session):
            self.session = session
        else:
            self.session = requests.Session()
        self.domain = domain
        self.headers = headers

        self.job_id = None

        # list of failed annotation imports
        self.failed_annotations = []

        # ready threading event
        self._ready = threading.Event()

    def create_job(self):
        """Create annotations import job through MD.ai API.
        This is an async operation. Status code of 202 indicates successful creation of job.
        """
        endpoint = f"https://{self.domain}/api/data-import/annotations"
        params = {
            "projectHashId": self.project_id,
            "datasetHashId": self.dataset_id,
            "annotations": self.annotations,
        }

        # reset list of failed annotation imports
        self.failed_annotations = []

        r = self.session.post(endpoint, json=params, headers=self.headers)

        if r.status_code == 202:
            self.job_id = r.json()["jobId"]
            msg = f"Importing {len(self.annotations)} annotations into "
            msg += f"project {self.project_id}, "
            msg += f"dataset {self.dataset_id}..."
            print(msg.ljust(100))
            self._check_job_progress()
        else:
            print(r.status_code)
            if r.status_code in (400, 401):
                msg = "Provided IDs are invalid, or you do not have sufficient permissions."
                print(msg)
            self._on_job_error()

    def wait_until_ready(self):
        self._ready.wait()

    @retry(
        retry_on_exception=retry_on_http_error,
        wait_exponential_multiplier=100,
        wait_exponential_max=1000,
        stop_max_attempt_number=10,
    )
    def _check_job_progress(self):
        """Poll for annotations import job progress.
        """
        endpoint = f"https://{self.domain}/api/data-import/annotations/progress"
        params = {"projectHashId": self.project_id, "jobId": self.job_id}
        r = self.session.post(endpoint, json=params, headers=self.headers)
        if r.status_code != 200:
            r.raise_for_status()

        try:
            body = r.json()
            status = body["status"]
        except (TypeError, KeyError):
            self._on_job_error()
            return

        if status == "done":
            self._on_job_done()

        elif status == "error":
            self._on_job_error()

        elif status == "running":
            try:
                progress = int(body["progress"])
            except (TypeError, ValueError):
                progress = 0
            try:
                time_remaining = int(body["timeRemaining"])
            except (TypeError, ValueError):
                time_remaining = 0

            # print formatted progress info
            if time_remaining > 45:
                time_remaining_fmt = (
                    arrow.now().shift(seconds=time_remaining).humanize(only_distance=True)
                )
            else:
                # arrow humanizes <= 45 to 'in seconds' or 'just now',
                # so we will opt to be explicit instead.
                time_remaining_fmt = f"{time_remaining} seconds"
            end_char = "\r" if progress < 100 else "\n"
            msg = (
                f"Annotations import for project {self.project_id}..."
                + f"{progress}% (time remaining: {time_remaining_fmt})."
            )
            print(msg.ljust(100), end=end_char, flush=True)

            # run progress check at 1s intervals so long as status == 'running'
            t = threading.Timer(1.0, self._check_job_progress)
            t.start()

    @retry(
        retry_on_exception=retry_on_http_error,
        wait_exponential_multiplier=100,
        wait_exponential_max=1000,
        stop_max_attempt_number=10,
    )
    def _on_job_done(self):
        endpoint = f"https://{self.domain}/api/data-import/annotations/done"
        params = {"projectHashId": self.project_id, "jobId": self.job_id}
        r = self.session.post(endpoint, json=params, headers=self.headers)
        if r.status_code != 200:
            r.raise_for_status()

        try:
            body = r.json()
            self.failed_annotations = body["failed"]
        except (TypeError, KeyError):
            self._on_job_error()
            return

        num_failed = len(self.failed_annotations)
        print(
            f"Successfully imported {len(self.annotations) - num_failed} / {len(self.annotations)}"
            + f" annotations into project {self.project_id}"
            + f", dataset {self.dataset_id}."
        )
        # fire ready threading.Event
        self._ready.set()

    @retry(
        retry_on_exception=retry_on_http_error,
        wait_exponential_multiplier=100,
        wait_exponential_max=1000,
        stop_max_attempt_number=10,
    )
    def _on_job_error(self):
        endpoint = f"https://{self.domain}/api/data-import/annotations/error"
        params = {"projectHashId": self.project_id, "jobId": self.job_id}
        r = self.session.post(endpoint, json=params, headers=self.headers)
        if r.status_code != 200:
            r.raise_for_status()
        print(
            f"Error importing annotations into project {self.project_id}"
            + f", dataset {self.dataset_id}."
        )
        # fire ready threading.Event
        self._ready.set()



import pydicom
import numpy as np
import colorsys
import random
import cv2
from skimage.measure import find_contours
import matplotlib.pyplot as plt
from matplotlib import patches


def random_colors(N, bright=True):
    """ Generate random colors.
    To get visually distinct colors, generate them in HSV space then convert to RGB.

    Args:
        N (int):
            Number of colors.
    """
    brightness = 1.0 if bright else 0.7
    hsv = [(i / N, 1, brightness) for i in range(N)]
    colors = list(map(lambda c: colorsys.hsv_to_rgb(*c), hsv))
    random.shuffle(colors)
    return colors


# based on functions in: https://github.com/matterport/Mask_RCNN/blob/master/mrcnn/visualize.py
def display_images(image_ids, titles=None, cols=3, cmap="gray", norm=None, interpolation=None):
    """Display images given image ids.

    Args:
        image_ids (list):
            List of image ids.

    TODO: figsize should not be hardcoded
    """
    titles = titles if titles is not None else [""] * len(image_ids)
    rows = len(image_ids) // cols + 1
    plt.figure(figsize=(14, 14 * rows // cols))
    i = 1
    for image_id, title in zip(image_ids, titles):
        plt.subplot(rows, cols, i)
        plt.title(title, fontsize=9)
        plt.axis("off")

        image = load_dicom_image(image_id, rescale=True)
        plt.imshow(image, cmap=cmap, norm=norm, interpolation=interpolation)

        i += 1
    plt.show()


def load_dicom_image(image_id, to_RGB=False, rescale=False):
    """ Load a DICOM image.

    Args:
        image_id (str):
            image id (filepath).
        to_RGB (bool, optional):
            Convert grayscale image to RGB.

    Returns:
        image array.
    """
    ds = pydicom.dcmread(image_id)
    try:
        image = ds.pixel_array
    except Exception:
        msg = (
            "Could not read pixel array from DICOM with TransferSyntaxUID "
            + ds.file_meta.TransferSyntaxUID
            + ". Likely unsupported compression format."
        )
        print(msg)

    if rescale:
        max_pixel_value = np.amax(image)
        min_pixel_value = np.amin(image)

        if max_pixel_value >= 255:
            # print("Input image pixel range exceeds 255, rescaling for visualization.")
            pixel_range = np.abs(max_pixel_value - min_pixel_value)
            pixel_range = pixel_range if pixel_range != 0 else 1
            image = image.astype(np.float32) / pixel_range * 255
            image = image.astype(np.uint8)

    if to_RGB:
        # If grayscale. Convert to RGB for consistency.
        if len(image.shape) != 3 or image.shape[2] != 3:
            image = np.stack((image,) * 3, -1)

    return image


def load_mask(image_id, dataset):
    """Load instance masks for the given image. Masks can be different types,
    mask is a binary true/false map of the same size as the image.

    """
    # annotations = imgs_anns[image_id]
    annotations = dataset.get_annotations_by_image_id(image_id)
    count = len(annotations)
    print("Number of annotations: %d" % count)

    image = load_dicom_image(image_id)
    width = image.shape[1]
    height = image.shape[0]

    if count == 0:
        print("No annotations")
        mask = np.zeros((height, width, 1), dtype=np.uint8)
        class_ids = np.zeros((1,), dtype=np.int32)
    else:
        mask = np.zeros((height, width, count), dtype=np.uint8)
        class_ids = np.zeros((count,), dtype=np.int32)

        for i, a in enumerate(annotations):

            label_id = a["labelId"]
            annotation_mode = dataset.label_id_to_class_annotation_mode(label_id)
            # print(annotation_mode)

            if annotation_mode == "bbox":
                # Bounding Box
                x = int(a["data"]["x"])
                y = int(a["data"]["y"])
                w = int(a["data"]["width"])
                h = int(a["data"]["height"])
                mask_instance = mask[:, :, i].copy()
                cv2.rectangle(mask_instance, (x, y), (x + w, y + h), 255, -1)
                mask[:, :, i] = mask_instance

            # FreeForm or Polygon
            elif annotation_mode == "freeform" or annotation_mode == "polygon":
                vertices = np.array(a["data"]["vertices"])
                vertices = vertices.reshape((-1, 2))
                mask_instance = mask[:, :, i].copy()
                cv2.fillPoly(mask_instance, np.int32([vertices]), (255, 255, 255))
                mask[:, :, i] = mask_instance

            # Line
            elif annotation_mode == "line":
                vertices = np.array(a["data"]["vertices"])
                vertices = vertices.reshape((-1, 2))
                mask_instance = mask[:, :, i].copy()
                cv2.polylines(mask_instance, np.int32([vertices]), False, (255, 255, 255), 12)
                mask[:, :, i] = mask_instance

            elif annotation_mode == "location":
                # Bounding Box
                x = int(a["data"]["x"])
                y = int(a["data"]["y"])
                mask_instance = mask[:, :, i].copy()
                cv2.circle(mask_instance, (x, y), 7, (255, 255, 255), -1)
                mask[:, :, i] = mask_instance

            elif annotation_mode == "mask":
                mask_instance = mask[:, :, i].copy()
                if a.data["foreground"]:
                    for i in a.data["foreground"]:
                        mask_instance = cv2.fillPoly(mask_instance, [np.array(i, dtype=np.int32)], (255, 255, 255))
                if a.data["background"]:
                    for i in a.data["background"]:
                        mask_instance = cv2.fillPoly(mask_instance, [np.array(i, dtype=np.int32)], (0,0,0))
                mask[:, :, i] = mask_instance

            elif annotation_mode is None:
                print("Not a local instance")

            # load class id
            class_ids[i] = dataset.label_id_to_class_id(label_id)

    return mask.astype(np.bool), class_ids.astype(np.int32)


def apply_mask(image, mask, color, alpha=0.3):
    """Apply the given mask to the image.

    Args:
        image: height, widht, channel.

    Returns:
        image with applied color mask.
    """
    for c in range(3):
        image[:, :, c] = np.where(
            mask == 1, image[:, :, c] * (1 - alpha) + alpha * color[c] * 255, image[:, :, c]
        )
    return image


def extract_bboxes(mask):
    """Compute bounding boxes from masks.

    Args:
        mask [height, width, num_instances]:
            Mask pixels are either 1 or 0.

    Returns:
        bounding box array [num_instances, (y1, x1, y2, x2)].
    """
    boxes = np.zeros([mask.shape[-1], 4], dtype=np.int32)
    for i in range(mask.shape[-1]):
        m = mask[:, :, i]
        # Bounding box.
        horizontal_indicies = np.where(np.any(m, axis=0))[0]
        vertical_indicies = np.where(np.any(m, axis=1))[0]
        if horizontal_indicies.shape[0]:
            x1, x2 = horizontal_indicies[[0, -1]]
            y1, y2 = vertical_indicies[[0, -1]]
            # x2 and y2 should not be part of the box. Increment by 1.
            x2 += 1
            y2 += 1
        else:
            # No mask for this instance. Might happen due to
            # resizing or cropping. Set bbox to zeros
            x1, x2, y1, y2 = 0, 0, 0, 0
        boxes[i] = np.array([y1, x1, y2, x2])
    return boxes.astype(np.int32)


def get_image_ground_truth(image_id, dataset):
    """Load and return ground truth data for an image (image, mask, bounding boxes).

    Args:
        image_id:
            Image id.

    Returns:
        image:
            [height, width, 3]
        class_ids:
            [instance_count] Integer class IDs
        bbox:
            [instance_count, (y1, x1, y2, x2)]
        mask:
            [height, width, instance_count]. The height and width are those of the image unless
            use_mini_mask is True, in which case they are defined in MINI_MASK_SHAPE.
    """
    # image = load_dicom_image(image_id, to_RGB=True)
    image = load_dicom_image(image_id, to_RGB=True, rescale=True)

    mask, class_ids = load_mask(image_id, dataset)

    _idx = np.sum(mask, axis=(0, 1)) > 0
    mask = mask[:, :, _idx]
    class_ids = class_ids[_idx]

    # Bounding boxes. Note that some boxes might be all zeros
    # if the corresponding mask got cropped out.
    # bbox: [num_instances, (y1, x1, y2, x2)]
    bbox = extract_bboxes(mask)

    return image, class_ids, bbox, mask


def display_annotations(
    image,
    boxes,
    masks,
    class_ids,
    scores=None,
    title="",
    figsize=(16, 16),
    ax=None,
    show_mask=True,
    show_bbox=True,
    colors=None,
    captions=None,
):
    """Display annotations for image.

    Args:
        boxes:
            [num_instance, (y1, x1, y2, x2, class_id)] in image coordinates.
        masks:
            [height, width, num_instances]
        class_ids:
            [num_instances]
        scores:
            (optional) confidence scores for each box
        title:
            (optional) Figure title
        show_mask, show_bbox:
            To show masks and bounding boxes or not
        figsize:
            (optional) the size of the image
        colors:
            (optional) An array or colors to use with each object
        captions:
            (optional) A list of strings to use as captions for each object
    """

    # Number of instancesload_mask
    N = boxes.shape[0]
    if not N:
        print("\n*** No instances to display *** \n")
    else:
        assert boxes.shape[0] == masks.shape[-1] == class_ids.shape[0]

    # If no axis is passed, create one and automatically call show()
    auto_show = False
    if not ax:
        _, ax = plt.subplots(1, figsize=figsize)
        auto_show = True

    # Generate random colors
    colors = colors or random_colors(N)

    # Show area outside image boundaries.
    height, width = image.shape[:2]
    ax.set_ylim(height + 10, -10)
    ax.set_xlim(-10, width + 10)
    ax.axis("off")
    ax.set_title(title)

    masked_image = image.astype(np.uint32).copy()
    for i in range(N):
        color = colors[i]

        # Bounding box
        if not np.any(boxes[i]):
            # Skip this instance. Has no bbox. Likely lost in image cropping.
            continue
        y1, x1, y2, x2 = boxes[i]
        if show_bbox:
            p = patches.Rectangle(
                (x1, y1),
                x2 - x1,
                y2 - y1,
                linewidth=2,
                alpha=0.7,
                linestyle="dashed",
                edgecolor=color,
                facecolor="none",
            )
            ax.add_patch(p)

        # Label
        if not captions:
            class_id = class_ids[i]
            score = scores[i] if scores is not None else None

            label = class_id
            caption = "{} {:.3f}".format(label, score) if score else label
        else:
            caption = captions[i]
        ax.text(x1, y1 + 8, caption, color="w", size=11, backgroundcolor="none")

        # Mask
        mask = masks[:, :, i]
        if show_mask:
            masked_image = apply_mask(masked_image, mask, color)

        # Mask Polygon
        # Pad to ensure proper polygons for masks that touch image edges.
        padded_mask = np.zeros((mask.shape[0] + 2, mask.shape[1] + 2), dtype=np.uint8)
        padded_mask[1:-1, 1:-1] = mask
        contours = find_contours(padded_mask, 0.5)
        for verts in contours:
            # Subtract the padding and flip (y, x) to (x, y)
            verts = np.fliplr(verts) - 1
            p = patches.Polygon(verts, facecolor="none", edgecolor=color)
            ax.add_patch(p)
    ax.imshow(masked_image.astype(np.uint8))
    # ax.imshow(masked_image)
    if auto_show:
        plt.show()


def draw_box_on_image(image, boxes, h, w):
    """Draw box on an image.

    Args:
        image:
            three channel (e.g. RGB) image.
        boxes:
            normalized box coordinate (between 0.0 and 1.0).
        h:
            image height
        w:
            image width
    """

    for i in range(len(boxes)):
        (left, right, top, bottom) = (
            boxes[i][0] * w,
            boxes[i][2] * w,
            boxes[i][1] * h,
            boxes[i][3] * h,
        )
        p1 = (int(left), int(top))
        p2 = (int(right), int(bottom))
        cv2.rectangle(image, p1, p2, (77, 255, 9), 3, 1)



"""MD.ai Python client library."""

try:
    from importlib import metadata
except ImportError:  # for Python<3.8
    import importlib_metadata as metadata
__version__ = metadata.version("mdai")

from . import preprocess
from .client import Client
from .utils import common_utils
from .utils import transforms

try:
    CAN_VISUALIZE = True
    from . import visualize
except ImportError:
    # matplotlib backend missing or cannot be loaded
    CAN_VISUALIZE = False



import os
import cv2
import numpy as np
import dicom2nifti
import pydicom

DEFAULT_IMAGE_SIZE = (512.0, 512.0)


def apply_slope_intercept(dicom_file):
    """
    Applies rescale slope and rescale intercept transformation.
    """
    array = dicom_file.pixel_array.copy()

    scale_slope = 1
    scale_intercept = 0
    if "RescaleIntercept" in dicom_file:
        scale_intercept = int(dicom_file.RescaleIntercept)
    if "RescaleSlope" in dicom_file:
        scale_slope = int(dicom_file.RescaleSlope)
    array = array * scale_slope
    array = array + scale_intercept
    return array


def remove_padding(array):
    """
    Removes background/padding from an 8bit numpy array.
    """
    arr = array.copy()
    nonzeros = np.nonzero(arr)
    x1 = np.min(nonzeros[0])
    x2 = np.max(nonzeros[0])
    y1 = np.min(nonzeros[1])
    y2 = np.max(nonzeros[1])
    return arr[x1:x2, y1:y2]


def get_window_from_dicom(dicom_file):
    """
    Returns window width and window center values.
    If no window width/level is provided or available, returns None.
    """
    width, level = None, None
    if "WindowWidth" in dicom_file:
        width = dicom_file.WindowWidth
        if isinstance(width, pydicom.multival.MultiValue):
            width = int(width[0])
        else:
            width = int(width)

    if "WindowCenter" in dicom_file:
        level = dicom_file.WindowCenter
        if isinstance(level, pydicom.multival.MultiValue):
            level = int(level[0])
        else:
            level = int(level)
    return width, level


def window(array, width, level):
    """
    Applies windowing operation.
    If window width/level is None, returns the array itself.
    """
    if width is not None and level is not None:
        array = np.clip(array, level - width // 2, level + width // 2)
    return array


def rescale_to_8bit(array):
    """
    Convert an array to 8bit (0-255).
    """
    array = array - np.min(array)
    array = array / np.max(array)
    array = (array * 255).astype("uint8")
    return array


def load_dicom_array(dicom_file, apply_slope_intercept=True):
    """
    Returns the dicom image as a Numpy array.
    """
    array = dicom_file.pixel_array.copy()
    if apply_slope_intercept:
        array = apply_slope_intercept(dicom_file)
    return array


def convert_dicom_to_nifti(dicom_files, tempdir):
    """
    Converts a dicom series to nifti format.
    Saves nifti in directory provided with filename as SeriesInstanceUID.nii.gz
    Returns a sorted list of dicom files based on image position patient.
    """
    output_file = os.path.join(tempdir, dicom_files[0].SeriesInstanceUID + ".nii.gz")
    nifti_file = dicom2nifti.convert_dicom.dicom_array_to_nifti(
        dicom_files, output_file=output_file, reorient_nifti=True,
    )
    return dicom2nifti.common.sort_dicoms(dicom_files)


def convert_dicom_to_8bit(dicom_file, imsize=None, width=None, level=None, keep_padding=True):
    """
    Given a DICOM file, window specifications, and image size,
    return the image as a Numpy array scaled to [0,255] of the specified size.
    """
    if width is None or level is None:
        width, level = get_window_from_dicom(dicom_file)

    array = apply_slope_intercept(dicom_file)
    array = window(array, width, level)
    array = rescale_to_8bit(array)

    if (
        "PhotometricInterpretation" in dicom_file
        and dicom_file.PhotometricInterpretation == "MONOCHROME1"
    ):
        array = 255 - array

    if not keep_padding:
        array = remove_padding(array)

    if imsize is not None:
        array = cv2.resize(array, imsize)
    return array


def convert_to_RGB(array, imsize=None):
    """
    Converts a single channel monochrome image to a 3 channel RGB image.
    """
    img = np.stack((array,) * 3, axis=-1)
    if imsize is not None:
        img = cv2.resize(img, imsize)
    return img


def convert_to_RGB_window(array, width, level, imsize=None):
    """
    Converts a monochrome image to 3 channel RGB with windowing.
    Width and level can be lists for different values per channel.
    """
    if type(width) is list and type(level) is list:
        R = window(array, width[0], level[0])
        G = window(array, width[1], level[1])
        B = window(array, width[2], level[2])
        img = np.stack([R, G, B], axis=-1)
    else:
        R = window(array, width, level)
        img = np.stack((R,) * 3, axis=-1)

    if imsize is not None:
        img = cv2.resize(img, imsize)
    return img


def stack_slices(dicom_files):
    """
    Stacks the +-1 slice to each slice in a dicom series.
    Returns the list of stacked images and sorted list of dicom files.
    """
    dicom_files = dicom2nifti.common.sort_dicoms(dicom_files)
    dicom_images = [load_dicom_array(i) for i in dicom_files]

    stacked_images = []
    for i, file in enumerate(dicom_images):
        if i == 0:
            img = np.stack([dicom_images[i], dicom_images[i], dicom_images[i + 1]], axis=-1)
            stacked_images.append(img)
        elif i == len(dicom_files) - 1:
            img = np.stack([dicom_images[i - 1], dicom_images[i], dicom_images[i]], axis=-1)
            stacked_images.append(img)
        else:
            img = np.stack([dicom_images[i - 1], dicom_images[i], dicom_images[i + 1]], axis=-1)
            stacked_images.append(img)

    return stacked_images, dicom_files



from PIL import Image
import tensorflow as tf
import sys
import io
import hashlib
from object_detection.utils import dataset_util
from mdai import visualize


def create_tf_bbox_example(annotations, image_id, classes_dict):

    image = visualize.load_dicom_image(image_id)
    width = int(image.shape[1])
    height = int(image.shape[0])

    raw_img = visualize.load_dicom_image(image_id, to_RGB=True)
    img = Image.fromarray(raw_img)
    img_buffer = io.BytesIO()
    img.save(img_buffer, format="jpeg")
    encoded_jpg = Image.open(img_buffer)

    if encoded_jpg.format != "JPEG":
        raise ValueError("Image format not JPEG")

    key = hashlib.sha256(img_buffer.getvalue()).hexdigest()

    xmins = []  # List of normalized left x coordinates in bounding box (1 per box)
    xmaxs = []  # List of normalized right x coordinates in bounding box (1 per box)
    ymins = []  # List of normalized top y coordinates in bounding box (1 per box)
    ymaxs = []  # List of normalized bottom y coordinates in bounding box (1 per box)
    classes_text = []  # List of string class name of bounding box (1 per box)
    classes = []  # List of integer class id of bounding box (1 per box)

    # per annotation
    for a in annotations:
        w = int(a["data"]["width"])
        h = int(a["data"]["height"])

        x_min = int(a["data"]["x"])
        y_min = int(a["data"]["y"])
        x_max = x_min + w
        y_max = y_min + h

        # WARN: these are normalized
        xmins.append(float(x_min / width))
        xmaxs.append(float(x_max / width))
        ymins.append(float(y_min / height))
        ymaxs.append(float(y_max / height))

        classes_text.append(a["labelId"].encode("utf8"))
        classes.append(classes_dict[a["labelId"]]["class_id"])

    # print(classes)

    tf_example = tf.train.Example(
        features=tf.train.Features(
            feature={
                "image/height": dataset_util.int64_feature(height),
                "image/width": dataset_util.int64_feature(width),
                "image/filename": dataset_util.bytes_feature(image_id.encode("utf8")),
                "image/source_id": dataset_util.bytes_feature(image_id.encode("utf8")),
                "image/key/sha256": dataset_util.bytes_feature(key.encode("utf8")),
                "image/encoded": dataset_util.bytes_feature(img_buffer.getvalue()),
                "image/format": dataset_util.bytes_feature("jpg".encode("utf8")),
                "image/object/bbox/xmin": dataset_util.float_list_feature(xmins),
                "image/object/bbox/xmax": dataset_util.float_list_feature(xmaxs),
                "image/object/bbox/ymin": dataset_util.float_list_feature(ymins),
                "image/object/bbox/ymax": dataset_util.float_list_feature(ymaxs),
                "image/object/class/text": dataset_util.bytes_list_feature(classes_text),
                "image/object/class/label": dataset_util.int64_list_feature(classes),
            }
        )
    )

    return tf_example


def write_to_tfrecords(output_path, dataset):
    """Write images and annotations to tfrecords.
    Args:
        output_path (str): Output file path of the TFRecord.
        dataset (object): Mdai dataset object.
    Examples:

        >>> train_record_fp = os.path.abspath('./train.record')
        >>> export.write_to_tfrecords(train_record_fp, train_dataset, label_ids_dict)
    """

    def _print_progress(count, total):
        # Percentage completion.
        pct_complete = float(count) / total

        # Status-message.
        # Note the \r which means the line should overwrite itself.
        msg = "\r- Progress: {0:.1%}".format(pct_complete)

        # Print it.
        sys.stdout.write(msg)
        sys.stdout.flush()

    print("\nOutput File Path: %s" % output_path)
    writer = tf.python_io.TFRecordWriter(output_path)
    num_images = len(dataset.image_ids)
    for i, image_id in enumerate(dataset.image_ids):
        _print_progress(count=i, total=num_images - 1)
        annotations = dataset.imgs_anns[image_id]
        tf_example = create_tf_bbox_example(annotations, image_id, dataset.classes_dict)
        writer.write(tf_example.SerializeToString())
    writer.close()



import random
import copy
import json
import pandas as pd

import os
import uuid
from functools import partial
import multiprocessing
import pydicom
import numpy as np
import nibabel as nib
from tqdm import tqdm
import cv2


def hex2rgb(h):
    """Convert Hex color encoding to RGB color"""
    h = h.lstrip("#")
    return tuple(int(h[i : i + 2], 16) for i in (0, 2, 4))


def train_test_split(dataset, shuffle=True, validation_split=0.1):
    """
    Split image ids into training and validation sets.
    TODO: Need to update dataset.dataset_data!
    TODO: What to set for images_dir for combined dataset?
    """
    if validation_split < 0.0 or validation_split > 1.0:
        raise ValueError(f"{validation_split} is not a valid split ratio.")

    image_ids_list = dataset.get_image_ids()
    if shuffle:
        sorted(image_ids_list)
        random.seed(42)
        random.shuffle(image_ids_list)

    split_index = int((1 - validation_split) * len(image_ids_list))
    train_image_ids = image_ids_list[:split_index]
    valid_image_ids = image_ids_list[split_index:]

    def filter_by_ids(ids, imgs_anns_dict):
        return {x: imgs_anns_dict[x] for x in ids}

    train_dataset = copy.deepcopy(dataset)
    train_dataset.id = dataset.id + "-TRAIN"
    train_dataset.name = dataset.name + "-TRAIN"

    valid_dataset = copy.deepcopy(dataset)
    valid_dataset.id = dataset.id + "-VALID"
    valid_dataset.name = dataset.name + "-VALID"

    imgs_anns_dict = dataset.imgs_anns_dict

    train_imgs_anns_dict = filter_by_ids(train_image_ids, imgs_anns_dict)
    valid_imgs_anns_dict = filter_by_ids(valid_image_ids, imgs_anns_dict)

    train_dataset.image_ids = train_image_ids
    valid_dataset.image_ids = valid_image_ids

    train_dataset.imgs_anns_dict = train_imgs_anns_dict
    valid_dataset.imgs_anns_dict = valid_imgs_anns_dict

    all_train_annotations = []
    for _, annotations in train_dataset.imgs_anns_dict.items():
        all_train_annotations += annotations
    train_dataset.all_annotations = all_train_annotations

    all_val_annotations = []
    for _, annotations in valid_dataset.imgs_anns_dict.items():
        all_val_annotations += annotations
    valid_dataset.all_annotations = all_val_annotations

    print(
        "Num of instances for training set: %d, validation set: %d"
        % (len(train_image_ids), len(valid_image_ids))
    )
    return train_dataset, valid_dataset


def json_to_dataframe(json_file, datasets=[]):
    with open(json_file, "r", encoding="utf-8") as f:
        data = json.load(f)

    a = pd.DataFrame([])
    studies = pd.DataFrame([])
    labels = None

    # Gets annotations for all datasets
    for d in data["datasets"]:
        if d["id"] in datasets or len(datasets) == 0:
            study = pd.DataFrame(d["studies"])
            study["dataset"] = d["name"]
            study["datasetId"] = d["id"]
            studies = studies.append(study, ignore_index=True, sort=False)

            annots = pd.DataFrame(d["annotations"])
            annots["dataset"] = d["name"]
            a = a.append(annots, ignore_index=True, sort=False)

    if len(studies) > 0:
        studies = studies[["StudyInstanceUID", "dataset", "datasetId", "number"]]
    g = pd.DataFrame(data["labelGroups"])
    # unpack arrays
    result = pd.DataFrame([(d, tup.id, tup.name) for tup in g.itertuples() for d in tup.labels])
    if len(result) > 0:
        result.columns = ["labels", "groupId", "groupName"]

        def unpack_dictionary(df, column):
            ret = None
            ret = pd.concat(
                [df, pd.DataFrame((d for idx, d in df[column].iteritems()))], axis=1, sort=False
            )
            del ret[column]
            return ret

        labels = unpack_dictionary(result, "labels")
        if "parentId" in labels.columns:
            labels = labels[
                [
                    "groupId",
                    "groupName",
                    "annotationMode",
                    "color",
                    "description",
                    "id",
                    "name",
                    "radlexTagIds",
                    "scope",
                    "parentId",
                ]
            ]
            labels.columns = [
                "groupId",
                "groupName",
                "annotationMode",
                "color",
                "description",
                "labelId",
                "labelName",
                "radlexTagIdsLabel",
                "scope",
                "parentLabelId",
            ]
        else:
            labels = labels[
                [
                    "groupId",
                    "groupName",
                    "annotationMode",
                    "color",
                    "description",
                    "id",
                    "name",
                    "radlexTagIds",
                    "scope",
                ]
            ]
            labels.columns = [
                "groupId",
                "groupName",
                "annotationMode",
                "color",
                "description",
                "labelId",
                "labelName",
                "radlexTagIdsLabel",
                "scope",
            ]

        if len(a) > 0:
            a = a.merge(labels, on=["labelId"], sort=False)
    if len(studies) > 0 and len(a) > 0:
        a = a.merge(studies, on=["StudyInstanceUID", "dataset"], sort=False)
    # Format data
    studies.number = studies.number.astype(int)
    a.number = a.number.astype(int)
    a.loc.createdAt = pd.to_datetime(a.createdAt)
    a.loc.updatedAt = pd.to_datetime(a.updatedAt)
    return {"annotations": a, "studies": studies, "labels": labels}


def convert_mask_annotation_to_array(row):
    """
    Converts a dataframe row containing a mask annotation from our internal complex polygon data representation to a numpy array.
    """
    mask = np.zeros((int(row.width), int(row.height)))
    if row.data["foreground"]:
        for i in row.data["foreground"]:
            mask = cv2.fillPoly(mask, [np.array(i, dtype=np.int32)], 1)
    if row.data["background"]:
        for i in row.data["background"]:
            mask = cv2.fillPoly(mask, [np.array(i, dtype=np.int32)], 0)
    return mask


def convert_mask_data(data):
    """
    Converts a numpy array mask to our internal complex polygon data representation.
    """
    mask = np.uint8(np.array(data) > 0)
    contours, hierarchy = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
    contours = [contours[i].reshape(-1, 2) for i in range(len(contours))]

    # Separate contours based on foreground / background polygons
    output_data = {
        "foreground": [],
        "background": [],
    }

    counts = [0] * len(contours)
    for i in range(len(contours)):
        parent = hierarchy[0][i][-1]
        if parent != -1:
            counts[i] = counts[parent] + 1

        if counts[i] % 2:
            output_data["background"].append(contours[i].tolist())
        else:
            output_data["foreground"].append(contours[i].tolist())
    return output_data


"""Converts NIFTI format to DICOM for CT exams. MR to come...

"""


def convert_ct(
    input_dir=None,
    output_dir=None,
    input_ext=".nii.gz",
    plane="axial",
    sample_dicom_fp=os.path.join(os.path.dirname(""), "./sample_dicom.dcm"),
    window_center=40,
    window_width=350,
):
    if not os.path.exists(input_dir):
        raise IOError("{:s} does not exist.".format(input_dir))
    if not os.path.exists(sample_dicom_fp):
        raise IOError("{:s} does not exist.".format(sample_dicom_fp))
    if plane not in ["axial", "sagittal", "coronal"]:
        raise ValueError("`plane` must be one of axial, sagittal, or coronal.")

    # make output dir if doesn't already exist
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    found_filepaths = list(_get_files(input_dir, ext=input_ext))
    print(f"{len(found_filepaths)} *{input_ext} files found. Processing...")

    n_procs = multiprocessing.cpu_count() - 1
    with multiprocessing.Pool(n_procs) as p:
        kwargs = {
            "input_dir": input_dir,
            "output_dir": output_dir,
            "input_ext": input_ext,
            "plane": plane,
            "sample_dicom_fp": sample_dicom_fp,
            "window_center": window_center,
            "window_width": window_width,
        }
        # need to iterate since Pool.imap is lazy
        for n in tqdm(
            p.imap_unordered(partial(_convert_nii_file_ct, **kwargs), found_filepaths),
            total=len(found_filepaths),
        ):
            pass


def _get_files(root, ext=None):
    """Yields all file paths recursively from root path, optionally filtering on extension.
    """
    for item in os.scandir(root):
        if item.is_file():
            if not ext:
                yield item.path
            elif os.path.splitext(item.path)[1] == ext or item.path.endswith(ext):
                yield item.path
        elif item.is_dir():
            yield from _get_files(item.path)


def _get_datatype(headers):
    dt = str(headers.get_data_dtype())
    return np.int16


# header datatype not reliable

#     if dt == 'int8':
#         return np.int8
#     elif dt == 'int16':
#         return np.int16
#     elif dt == 'int32':
#         return np.int32
#     elif dt == 'int64':
#         return np.int64
#     elif dt == 'float32':
#         return np.float32
#     elif dt == 'float64':
#         return np.float64
#     return np.int16


def _convert_nii_file_ct(
    filepath,
    input_dir=None,
    output_dir=None,
    input_ext=".nii.gz",
    plane="axial",
    sample_dicom_fp=os.path.join(os.path.dirname(""), "./sample_dicom.dcm"),
    window_center=40,
    window_width=350,
):
    dataobj = nib.load(filepath)
    headers = dataobj.header
    voxel_arr = dataobj.get_fdata()
    pixdim = headers["pixdim"][1:4].tolist()

    # NIFTI (RAS) -> DICOM (LPI) coordinates
    # i, Left/Right = sagittal plane
    # j, Anterior/Posterior = coronal plane
    # k, Superior/Inferior = axial plane
    voxel_arr = np.flip(voxel_arr, 0)
    voxel_arr = np.flip(voxel_arr, 1)
    voxel_arr = np.flip(voxel_arr, 2)

    # Image coordinates -> World coordinates
    if plane == "axial":
        slice_axis = 2
        plane_axes = [0, 1]
    elif plane == "coronal":
        slice_axis = 1
        plane_axes = [0, 2]
    elif plane == "sagittal":
        slice_axis = 0
        plane_axes = [1, 2]
    thickness = pixdim[slice_axis]
    spacing = [pixdim[plane_axes[1]], pixdim[plane_axes[0]]]
    voxel_arr = np.swapaxes(voxel_arr, *plane_axes)

    # generate DICOM UIDs (StudyInstanceUID and SeriesInstanceUID)
    study_uid = pydicom.uid.generate_uid(prefix=None)
    series_uid = pydicom.uid.generate_uid(prefix=None)

    # randomized patient ID
    patient_id = str(uuid.uuid4())
    patient_name = patient_id

    try:
        scale_slope = str(int(headers["scl_slope"]))
    except ValueError:  # handle NaN
        scale_slope = "1"
    try:
        scale_intercept = str(int(headers["scl_inter"]))
    except ValueError:  # handle NaN
        scale_intercept = "0"

    for slice_index in range(voxel_arr.shape[slice_axis]):
        # generate SOPInstanceUID
        instance_uid = pydicom.uid.generate_uid(prefix=None)

        loc = slice_index * thickness

        ds = pydicom.dcmread(sample_dicom_fp)

        # delete tags
        del ds[0x00200052]  # Frame of Reference UID
        del ds[0x00201040]  # Position Reference Indicator

        # slice and set PixelData tag
        axes = [slice(None)] * 3
        axes[slice_axis] = slice_index
        arr = voxel_arr[tuple(axes)].astype(_get_datatype(headers))
        ds[0x7FE00010].value = arr.tobytes()

        # modify tags
        # - UIDs are created by pydicom.uid.generate_uid at each level above
        # - image position is calculated by combination of slice index and slice thickness
        # - slice location is set to the value of image position along z-axis
        # - Rows/Columns determined by array shape
        # - we set slope/intercept to 1/0 since we're directly converting from PNG pixel values
        ds[0x00080018].value = instance_uid  # SOPInstanceUID
        ds[0x00100010].value = patient_name
        ds[0x00100020].value = patient_id
        ds[0x0020000D].value = study_uid  # StudyInstanceUID
        ds[0x0020000E].value = series_uid  # SeriesInstanceUID
        ds[0x0008103E].value = ""  # Series Description
        ds[0x00200011].value = "1"  # Series Number
        ds[0x00200012].value = str(slice_index + 1)  # Acquisition Number
        ds[0x00200013].value = str(slice_index + 1)  # Instance Number
        ds[0x00201041].value = str(loc)  # Slice Location
        ds[0x00280010].value = arr.shape[0]  # Rows
        ds[0x00280011].value = arr.shape[1]  # Columns
        ds[0x00280030].value = spacing  # Pixel Spacing
        ds[0x00281050].value = str(window_center)  # Window Center
        ds[0x00281051].value = str(window_width)  # Window Width
        ds[0x00281052].value = str(scale_intercept)  # Rescale Intercept
        ds[0x00281053].value = str(scale_slope)  # Rescale Slope

        # Image Position (Patient)
        # Image Orientation (Patient)
        if plane == "axial":
            ds[0x00200032].value = ["0", "0", str(loc)]
            ds[0x00200037].value = ["1", "0", "0", "0", "1", "0"]
        elif plane == "coronal":
            ds[0x00200032].value = ["0", str(loc), "0"]
            ds[0x00200037].value = ["1", "0", "0", "0", "0", "1"]
        elif plane == "sagittal":
            ds[0x00200032].value = [str(loc), "0", "0"]
            ds[0x00200037].value = ["0", "1", "0", "0", "0", "1"]

        # add new tags
        # see tag info e.g., from https://dicom.innolitics.com/ciods/nm-image/nm-reconstruction/00180050
        # Slice Thickness
        ds[0x00180050] = pydicom.dataelem.DataElement(0x00180050, "DS", str(thickness))

        # Output DICOM filepath
        # For root directory of data/, then:
        # e.g., 'data/x/y/z.nii.gz' becomes '{output_dir}/data/x/y/z/{001-999}.dcm'
        dicom_fp = os.path.join(
            output_dir,
            os.path.dirname(filepath).strip("/"),  # remove leading and trailing slashes
            os.path.basename(filepath).replace(input_ext, ""),
            "{:03}.dcm".format(slice_index + 1),
        )

        # create directory
        if not os.path.exists(os.path.dirname(dicom_fp)):
            os.makedirs(os.path.dirname(dicom_fp))

        # write DICOM to file
        pydicom.dcmwrite(dicom_fp, ds)



from mdai.visualize import load_dicom_image
from keras.utils import Sequence, to_categorical

import numpy as np
from PIL import Image


class DataGenerator(Sequence):
    def __init__(
        self,
        dataset,
        batch_size=32,
        dim=(32, 32),
        n_channels=1,
        n_classes=10,
        shuffle=True,
        to_RGB=True,
        rescale=False,
    ):
        """Generates data for Keras fit_generator() function.
        """

        # Initialization
        self.dim = dim
        self.batch_size = batch_size

        self.img_ids = dataset.image_ids
        self.imgs_anns_dict = dataset.imgs_anns_dict
        self.dataset = dataset

        self.n_channels = n_channels
        self.n_classes = n_classes
        self.shuffle = shuffle
        self.to_RGB = to_RGB
        self.rescale = rescale
        self.on_epoch_end()

    def __len__(self):
        "Denotes the number of batches per epoch"
        return int(np.floor(len(self.img_ids) / self.batch_size))

    def __getitem__(self, index):
        "Generate one batch of data"

        # Generate indexes of the batch
        indexes = self.indexes[index * self.batch_size : (index + 1) * self.batch_size]

        # Find list of IDs
        img_ids_temp = [self.img_ids[k] for k in indexes]

        # Generate data
        X, y = self.__data_generation(img_ids_temp)

        return X, y

    def on_epoch_end(self):
        "Updates indexes after each epoch"
        self.indexes = np.arange(len(self.img_ids))
        if self.shuffle:
            np.random.shuffle(self.indexes)

    def __data_generation(self, img_ids_temp):
        "Generates data containing batch_size samples"

        # Initialization
        X = np.empty((self.batch_size, *self.dim, self.n_channels))
        y = np.empty((self.batch_size), dtype=int)

        # Generate data
        for i, ID in enumerate(img_ids_temp):
            image = load_dicom_image(ID, to_RGB=self.to_RGB, rescale=self.rescale)
            try:
                image = Image.fromarray(image)
            except Exception:
                print(
                    "Pil.Image can't read image. Possible 12 or 16 bit image. Try rescale=True to "
                    + "scale to 8 bit."
                )

            image = image.resize((self.dim[0], self.dim[1]))

            X[i,] = image

            ann = self.imgs_anns_dict[ID][0]
            y[i] = self.dataset.classes_dict[ann["labelId"]]["class_id"]
        return X, to_categorical(y, num_classes=self.n_classes)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mdai', 'mdai.utils']

package_data = \
{'': ['*']}

install_requires = \
['arrow',
 'dicom2nifti',
 'matplotlib',
 'nibabel',
 'numpy',
 'opencv-python',
 'pandas',
 'pillow',
 'pydicom',
 'requests',
 'retrying',
 'scikit-image',
 'tqdm']

setup_kwargs = {
    'name': 'mdai',
    'version': '0.11.0',
    'description': 'MD.ai Python client library',
    'long_description': '# MD.ai Python Client Library\n\n**Currently pre-alpha -- API may change significantly in future releases.**\n\nThe python client library is designed to work with the datasets and annotations generated by the [MD.ai](https://www.md.ai/) Medical AI platform.\n\nYou can download datasets consisting of images and annotations (i.e., JSON file), create train/validation/test datasets, integrate with various machine learing libraries (e.g., TensorFlow/Keras, Fast.ai) for developing machine learning algorithms.\n\nTo get started, check out the examples in the [notebooks section](notebooks), or our [intro to deep learning for medical imaging lessons](https://github.com/mdai/ml-lessons/).\n\n## Installation\n\nRequires Python 3.6+. Install and update using [pip](https://pip.pypa.io/en/stable/quickstart/):\n\n```sh\npip install --upgrade mdai\n```\n\n## Documentation\n\nOverview Documentation: https://docs.md.ai/libraries/python/\nAPI Documentation coming soon.\n\n## The MD.ai Annotator\n\nThe MD.ai annotator is a powerful web based application, to store and view anonymized medical images (e.g, DICOM) on the cloud, create annotations collaboratively, in real-time, and export annotations, images and labels for training. The MD.ai python client library can be used to download images and annotations, prepare the datasets, and then used to train and evaluate deep learning models.\n\n- MD.ai Documentation and Videos URL: https://docs.md.ai/\n- MD.ai Annotator Example Project URL: https://public.md.ai/annotator/project/aGq4k6NW/workspace\n\n![MD.ai Annotator](https://docs.md.ai/img/annotator-homepage.png)\n\n## MD.ai Annotation JSON Format\n\nMore detailed information regarding the annotation JSON export format, see: https://docs.md.ai/data/json/\n\n## Example Notebooks\n\n- [HelloWorld Keras Notebook](notebooks/hello-world-keras.ipynb)\n- [HelloWorld TFRecords Notebook](notebooks/hello-world-tfrecords-VGG16.ipynb)\n- [HelloWorld Fast.ai](notebooks/hello-world-fastai.ipynb)\n\n## Introductory lessons to Deep Learning for medical imaging by [MD.ai](https://www.md.ai)\n\nThe following are several Jupyter notebooks covering the basics of downloading and parsing annotation data, and training and evaluating different deep learning models for classification, semantic and instance segmentation and object detection problems in the medical imaging domain. The notebooks can be run on Google\'s colab with GPU (see instruction below).\n\n- Lesson 1. Classification of chest vs. adominal X-rays using TensorFlow/Keras [Github](https://github.com/mdai/ml-lessons/blob/master/lesson1-xray-images-classification.ipynb) | [Annotator](https://public.md.ai/annotator/project/PVq9raBJ)\n- Lesson 2. Lung X-Rays Semantic Segmentation using UNets. [Github](https://github.com/mdai/ml-lessons/blob/master/lesson2-lung-xrays-segmentation.ipynb) |\n  [Annotator](https://public.md.ai/annotator/project/aGq4k6NW/workspace)\n- Lesson 3. RSNA Pneumonia detection using Kaggle data format [Github](https://github.com/mdai/ml-lessons/blob/master/lesson3-rsna-pneumonia-detection-kaggle.ipynb) | [Annotator](https://public.md.ai/annotator/project/LxR6zdR2/workspace)\n- Lesson 3. RSNA Pneumonia detection using MD.ai python client library [Github](https://github.com/mdai/ml-lessons/blob/master/lesson3-rsna-pneumonia-detection-mdai-client-lib.ipynb) | [Annotator](https://public.md.ai/annotator/project/LxR6zdR2/workspace)\n\n## Contributing\n\nSee [contributing guidelines](CONTRIBUTING.md) to set up a development environemnt and how to make contributions to mdai.\n\n## Running Jupyter notebooks Colab\n\nIt’s easy to run a Jupyter notebook on Google\'s Colab with free GPU use (time limited).\nFor example, you can add the Github jupyter notebook path to https://colab.research.google.com/notebook:\nSelect the "GITHUB" tab, and add the Lesson 1 URL: https://github.com/mdai/ml-lessons/blob/master/lesson1-xray-images-classification.ipynb\n\nTo use the GPU, in the notebook menu, go to Runtime -> Change runtime type -> switch to Python 3, and turn on GPU. See more [colab tips.](https://www.kdnuggets.com/2018/02/essential-google-colaboratory-tips-tricks.html)\n\n## Advanced: How to run on Google Cloud Platform with Deep Learning Images\n\nYou can also run the notebook with powerful GPUs on the Google Cloud Platform. In this case, you need to authenticate to the Google Cloug Platform, create a private virtual machine instance running a Google\'s Deep Learning image, and import the lessons. See instructions below.\n\n[GCP Deep Learnings Images How To](running_on_gcp.md)\n\n---\n\n&copy; 2022 MD.ai, Inc.\n',
    'author': 'MD.ai',
    'author_email': 'github@md.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mdai/mdai-client-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



import os
import json
import collections
import glob


class Project:
    """Project consists of label groups, and datasets.

    Args:
        annotations_fp (str):
            File path to the exported JSON annotation file.
        images_dir (str):
            File path to the DICOM images directory.
    """

    def __init__(self, annotations_fp=None, images_dir=None):
        """

        """
        self.annotations_fp = None
        self.images_dir = None
        self.label_groups = []
        self.datasets = []

        if annotations_fp is not None and images_dir is not None:
            self.annotations_fp = annotations_fp
            self.images_dir = images_dir

            with open(self.annotations_fp, "r") as f:
                self.data = json.load(f)

            for dataset in self.data["datasets"]:
                self.datasets.append(Dataset(dataset, images_dir))

            for label_group in self.data["labelGroups"]:
                self.label_groups.append(LabelGroup(label_group))
        else:
            print("Error: Missing data or images file paths!")

    def get_label_groups(self):
        return self.label_groups

    def show_label_groups(self):
        for label_group in self.label_groups:
            print("Label Group, Id: %s, Name: %s" % (label_group.id, label_group.name))
            label_group.show_labels("\t")

    def get_label_group_by_name(self, label_group_name):
        for label_group in self.label_groups:
            if label_group.name == label_group_name:
                return label_group
        return None

    def get_label_group_by_id(self, label_group_id):
        for label_group in self.label_groups:
            if label_group.id == label_group_id:
                return label_group
        return None

    def get_datasets(self):
        """Get JSON representation of datasets"""
        return self.datasets

    def show_datasets(self):
        print("Datasets:")
        for dataset in self.datasets:
            print("Id: %s, Name: %s" % (dataset.id, dataset.name))
        print("")

    def get_dataset_by_name(self, dataset_name):
        for dataset in self.datasets:
            if dataset.name == dataset_name:
                return dataset
        raise ValueError(f"Dataset name {dataset_name} does not exist.")

    def get_dataset_by_id(self, dataset_id):
        for dataset in self.datasets:
            if dataset.id == dataset_id:
                return dataset
        raise ValueError(f"Dataset id {dataset_id} does not exist.")

    def set_labels_dict(self, labels_dict):

        self.classes_dict = self._create_classes_dict(labels_dict)

        for dataset in self.datasets:
            dataset.classes_dict = self.classes_dict

    def get_label_id_annotation_mode(self, label_id):
        "Return label id's annotation mode."
        for label_group in self.label_groups:
            labels_data = label_group.get_data()["labels"]
            for label in labels_data:
                if label["id"] == label_id:
                    return label["annotationMode"]
        raise ValueError(f"Label id {label_id} does not exist.")

    def get_label_id_type(self, label_id):
        "Return label id's type."
        for label_group in self.label_groups:
            labels_data = label_group.get_data()["labels"]
            for label in labels_data:
                if label["id"] == label_id:
                    return label["type"]
        raise ValueError(f"Label id {label_id} does not exist.")

    def get_label_id_scope(self, label_id):
        "Return label id's scope."
        for label_group in self.label_groups:
            labels_data = label_group.get_data()["labels"]
            for label in labels_data:
                if label["id"] == label_id:
                    return label["scope"]
        raise ValueError(f"Label id {label_id} does not exist.")

    def _create_classes_dict(self, labels_dict):
        """Create a dict with label id as key, and a nested dict of class_id, and class_text as \
        values, e.g., {'L_v8n': {'class_id': 1, 'class_text': 'Lung Opacity'}}, where L_v8n is \
        the label id, with a class_id of 1 and class text of 'Lung Opacity'.

        Args:
            labels_dict:
                dictionary containing label ids, and (user defined) class ids

        Returns:
            classes dict
        """
        classes_dict = {}

        for label_id, class_id in labels_dict.items():
            for label_group in self.label_groups:
                labels_data = label_group.get_data()["labels"]
                for label in labels_data:
                    if label["id"] == label_id:
                        if class_id == 0 and label["type"].lower() == "local":
                            raise Exception(
                                f"{label_id} is a local type, its class id cannot be 0."
                            )
                        classes_dict[label_id] = {
                            "class_id": class_id,
                            "class_text": label["name"],
                            "class_annotation_mode": label["annotationMode"],
                            "scope": label["scope"],
                            "type": label["type"],
                        }

        if classes_dict.keys() != labels_dict.keys():
            in_labels = labels_dict.keys()
            out_labels = classes_dict.keys()
            diff = set(in_labels).symmetric_difference(out_labels)
            raise ValueError(f"Labels {diff} are not valid for this dataset.")

        return classes_dict


class LabelGroup:
    """A label group contains multiple labels.
    Each label has properties such id, name, color, type, scope, annotation mode, rad lex tag ids.

    Label type:
        Global typed annotations apply to the whole instance (e.g., a CT image), while
        local typed annotations apply to a part of the image (e.g., ROI bounding box).
    Label scope:
        Scope can be of study, series, or instance.
    Label annotation mode:
        Annotation mode can be of bounding boxes, free form, polygon, etc.
    """

    def __init__(self, label_group_data):
        """
        Args:
            label_group (object: json) JSON data for label group
        """
        self.label_group_data = label_group_data
        self.name = self.label_group_data["name"]
        self.id = self.label_group_data["id"]

    def get_data(self):
        return self.label_group_data

    def get_labels(self):
        """Get label ids and names """
        return [(label["id"], label["name"]) for label in self.label_group_data["labels"]]

    def show_labels(self, print_offset=""):
        """Show labels info"""
        print(f"{print_offset}Labels:")
        for label in self.label_group_data["labels"]:
            print(f"{print_offset}Id: {label['id']}, Name: {label['name']}")
        print("")


class Dataset:
    """A dataset consists of DICOM images and annotations.

    Args:
        dataset_data:
            Dataset json data.
        images_dir:
            DICOM images directory.
    """

    def __init__(self, dataset_data, images_dir):

        self.dataset_data = dataset_data
        self.images_dir = images_dir

        self.id = dataset_data["id"]
        self.name = dataset_data["name"]
        self.all_annotations = dataset_data["annotations"]

        self.image_ids = None
        self.classes_dict = None
        self.imgs_anns_dict = None

        # all image ids
        self.all_image_ids = glob.glob(os.path.join(self.images_dir, "**/*.dcm"), recursive=True)

    def prepare(self):
        if self.classes_dict is None:
            raise Exception("Use `Project.set_labels_dict()` to set labels.")

        label_ids = self.classes_dict.keys()

        # filter annotations by label ids
        ann_filtered = self.get_annotations(label_ids)
        if not ann_filtered:
            raise Exception(f"No annotations exist for dataset '{self.name}'.")

        self.imgs_anns_dict = self._associate_images_and_annotations(ann_filtered)

    def get_annotations(self, label_ids=None, verbose=False):
        """Returns annotations, filtered by label ids.

        Args:
            label_ids (optional):
                Filter returned annotations by matching label ids.

            verbose (optional:
                Print debug messages.
        """
        if label_ids is None:
            if verbose:
                print("Dataset contains %d annotations." % len(self.all_annotations))
            return self.all_annotations

        ann_filtered = [a for a in self.all_annotations if a["labelId"] in label_ids]

        if verbose:
            print(
                f"Dataset contains {len(ann_filtered)} annotations"
                + f", filtered by label ids {label_ids}."
            )
        return ann_filtered

    def _generate_uid(self, ann):
        """Generate an unique image identifier based on the DICOM file structure.

        Args:
            ann (list):
                List of annotations.

        Returns:
            A unique image identifier based on the DICOM file structure.
        """

        uid = None

        if "StudyInstanceUID" and "SeriesInstanceUID" and "SOPInstanceUID" in ann:
            # SOPInstanceUID aka image level
            uid = os.path.join(
                self.images_dir,
                ann["StudyInstanceUID"],
                ann["SeriesInstanceUID"],
                ann["SOPInstanceUID"] + ".dcm",
            )
            return uid
        elif "StudyInstanceUID" and "SeriesInstanceUID" in ann:
            prefix = os.path.join(
                self.images_dir, ann["StudyInstanceUID"], ann["SeriesInstanceUID"]
            )
            uid = [image_id for image_id in self.all_image_ids if image_id.startswith(prefix)]
            return uid
        elif "StudyInstanceUID" in ann:
            prefix = os.path.join(self.images_dir, ann["StudyInstanceUID"])
            uid = [image_id for image_id in self.all_image_ids if image_id.startswith(prefix)]
            return uid
        else:
            raise ValueError(f"Unable to create UID from {ann}")

    def get_image_ids(self, verbose=False):
        """Returns image ids. Must call prepare() method first in order to generate image ids.

        Args:
            verbose (Optional):
                Print debug message.
        """
        if not self.image_ids:
            raise Exception("Call project.prepare() first.")

        if verbose:
            print(
                f"Dataset contains {len(self.image_ids)} images"
                + f", filtered by label ids {self.classes_dict.keys()}."
            )
        return self.image_ids

    def _generate_image_ids(self, anns):
        """Get images ids for annotations.

        Args:
            ann (list):
            List of image ids.

        Returns:
            A list of image ids.
        """
        image_ids = set()
        for ann in anns:
            uid = self._generate_uid(ann)

            if uid:
                if isinstance(uid, list):
                    for one_uid in uid:
                        image_ids.add(one_uid)
                else:
                    image_ids.add(uid)

        # image_ids = glob.glob(os.path.join(self.images_dir, "**/*.dcm"), recursive=True)
        return sorted(list(image_ids))

    def get_annotations_by_image_id(self, image_id):
        if image_id not in self.image_ids:
            raise ValueError(f"Image id {image_id} is not found in dataset {self.name}.")

        return self.imgs_anns_dict[image_id]

    def _associate_images_and_annotations(self, anns):
        """Generate image ids to annotations mapping.
        Each image can have zero or more annotations.

        Args:
            anns (list):
                List of annotations.

        Returns:
            A dictionary with image ids as keys and annotations as values.
        """
        self.image_ids = self._generate_image_ids(anns)

        # empty dictionary with image ids as keys
        imgs_anns_dict = collections.OrderedDict()
        imgs_anns_dict = {fp: [] for fp in self.image_ids}

        for ann in anns:
            uid = self._generate_uid(ann)
            if uid:
                if isinstance(uid, list):
                    for one_uid in uid:
                        imgs_anns_dict[one_uid].append(ann)
                else:
                    imgs_anns_dict[uid].append(ann)

        return imgs_anns_dict

    def class_id_to_class_text(self, class_id):
        for k, v in self.classes_dict.items():
            if v["class_id"] == class_id:
                return v["class_text"]

        raise Exception(f"class_id {class_id} is invalid.")

    def class_text_to_class_id(self, class_text):
        for k, v in self.classes_dict.items():
            if v["class_text"] == class_text:
                return v["class_id"]
        raise Exception(f"class_text {class_text} is invalid.")

    def label_id_to_class_id(self, label_id):
        for k, v in self.classes_dict.items():
            if k == label_id:
                return v["class_id"]
        raise Exception(f"label_id {label_id} is invalid.")

    def label_id_to_class_annotation_mode(self, label_id):
        for k, v in self.classes_dict.items():
            if k == label_id:
                return v["class_annotation_mode"]
        raise Exception(f"label_id {label_id} is invalid.")

    def show_classes(self):
        for k, v in self.classes_dict.items():
            print(f"Label id: {k}, Class id: {v['class_id']}, Class text: {v['class_text']}")
